//  
//  Copyright (C) 2013 Michael Langfermann
// 
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
// 

public class AppDialog : Gtk.Dialog {
    private Gtk.ListStore list_store;
    private Gtk.TreeModelSort list_sort;
    private Gtk.TreeModelFilter list_filter;
    private Gtk.TreeView list;
    private Gtk.Widget apply_button;
    private Granite.Widgets.SearchBar search;
    private Gtk.ScrolledWindow scroll;
    private Gtk.CellRendererText cell;
    private Gtk.Entry command;

    private AppInfo buffer_ignore;
    private string search_text = "";

    private const int ICON_SIZE = 48;
    private const int[] WINDOW_SIZE = { 350 , 400 };
    private const string FALLBACK_ICON = "application-default-icon";

    public AppInfo selected_app;

    public string list_markup = "<span font_weight=\"bold\" size=\"large\">{0}</span>\n{1}";

    public bool custom_command { get; set; default = true;}

    public signal void app_chosen ();
    public signal void app_selected ();

    public AppDialog () {
        title = _("Choose an Application...");
        border_width = 5;
        set_default_size (WINDOW_SIZE[0], WINDOW_SIZE[1]);
        create_widgets ();
        list_update ();
        connect_signals ();
    }

    private void create_widgets () {
        // Create ListStore
        list_store = new Gtk.ListStore (3,
                                        typeof (string),        // Application name and description
                                        typeof (Gdk.Pixbuf),    // Application icon
                                        typeof (AppInfo));      // Hidden returned application data

        // Create a filter model wrapper for the search
        list_filter = new Gtk.TreeModelFilter (list_store, null);
        list_filter.set_visible_func (filter_func);

        // Create a sort model wrapper, sort in ascending order
        list_sort = new Gtk.TreeModelSort.with_model (list_filter);
        list_sort.set_sort_column_id (0, Gtk.SortType.ASCENDING);

        // Create a TreeView with the sorted and filtered model
        list = new Gtk.TreeView.with_model (list_sort);
        list.expand = true;
        list.enable_search = false;
        list.headers_visible = false;

        // Create and append columns
        Gtk.CellRendererPixbuf pixbuf = new Gtk.CellRendererPixbuf ();
        list.insert_column_with_attributes (-1, null, pixbuf, "pixbuf", 1);

        cell = new Gtk.CellRendererText ();
        cell.wrap_mode = Pango.WrapMode.WORD_CHAR;
        cell.wrap_width = WINDOW_SIZE[0] - (WINDOW_SIZE[0] / 3);
        list.insert_column_with_attributes (-1, null, cell, "markup", 0);

        // Search widget
        search = new Granite.Widgets.SearchBar (_("Search Applications..."));

        // Layout widgets
        scroll = new Gtk.ScrolledWindow (null, null);
        scroll.hscrollbar_policy = Gtk.PolicyType.NEVER;
        scroll.vscrollbar_policy = Gtk.PolicyType.AUTOMATIC;
        scroll.shadow_type = Gtk.ShadowType.IN;
        scroll.expand = true;

        scroll.add (list);

        Gtk.Box content = get_content_area () as Gtk.Box;
        content.pack_start (search, false, true, 0);
        content.pack_start (scroll, false, true, 0);

        command = new Gtk.Entry();
        command.placeholder_text = _("Type in a custom command");
        command.primary_icon_name = "document-properties-symbolic";
        command.primary_icon_activatable = false;
        
        if (custom_command == true)
            content.pack_start (command, false, true, 0);


        content.spacing = 10;

        // Add buttons to button area
        add_button (Gtk.Stock.CANCEL, Gtk.ResponseType.CLOSE);
        apply_button = add_button (Gtk.Stock.ADD, Gtk.ResponseType.APPLY);
        apply_button.sensitive = false;
    }

    // Load applications into ListStore
    private void list_update () {
        var apps = AppInfo.get_all ();
        var icon_theme = Gtk.IconTheme.get_default ();

        list_store.clear ();

        foreach (AppInfo app in apps) {
            if (app.should_show ()) {
                Gtk.TreeIter iter;
                var app_name = app.get_name ();
                var app_desc = app.get_description ();

                (app_desc == null)? app_desc = "":null;

                list_store.append (out iter);
                list_store.set (iter, 0, list_markup.replace("{0}", app_name)
                                                    .replace("{1}", app_desc));

                try {
                    if (Gtk.IconTheme.get_default ().has_icon (app.get_icon ().to_string ()))
                        list_store.set (iter, 1,
                                        icon_theme.load_icon (app.get_icon ().to_string (),
                                        ICON_SIZE, Gtk.IconLookupFlags.GENERIC_FALLBACK));
                    else
                        list_store.set (iter, 1,
                                        icon_theme.load_icon (FALLBACK_ICON,
                                        ICON_SIZE, Gtk.IconLookupFlags.GENERIC_FALLBACK));
                } catch (Error e) { warning (e.message); }

                list_store.set (iter, 2, app);
            }
        }
    }


    // Connect signals
    private void connect_signals () {
        // Set add button to sensitive when an application is highlighted
        list.cursor_changed.connect (() => {
            apply_button.sensitive = true;
            get_app();
            if (buffer_ignore != selected_app) {
                app_selected();
            }
            buffer_ignore = selected_app;
        });

        command.icon_press.connect (() => command.text = "");

        command.changed.connect (() => {
            var is_empty = (command.text == "");

            scroll.sensitive = is_empty;
            search.sensitive = is_empty;
            command.secondary_icon_sensitive = !is_empty;
            apply_button.sensitive = !is_empty;
            if (is_empty) {
                command.secondary_icon_name = null;
                list_update ();
            } else {
                list_store.clear ();
                command.secondary_icon_name = "edit-clear-symbolic";
            }
        });

        list.row_activated.connect (() => command.activate());

        // Connect the buttons
        response.connect (on_response);

        command.activate.connect (() => {
            get_app();
            app_chosen();
            destroy ();
        });

        // Connect the SearchBar to the filter
        search.text_changed_pause.connect ((text) => {
            search_text = text;
            list_filter.refilter();
        });

        // When the window gets resized, the application names and descriptions wrap accordingly
        size_allocate.connect ((allocation) =>
                               cell.wrap_width = allocation.width - (allocation.width / 3));
    }

    // Connect the buttons
    private void on_response (Gtk.Dialog source, int response_id) {
        switch (response_id) {
            case Gtk.ResponseType.APPLY:
                get_app();
                app_chosen();
                destroy ();
            break;
            case Gtk.ResponseType.CLOSE:
                destroy ();
            break;
        }
    }

    // Sets the selected app
    private void get_app () {
        Gtk.TreeIter iter;
        Gtk.TreePath path;
        AppInfo app = null;

        list.get_cursor (out path, null);
        if (path != null) {
            path =  list_sort.convert_path_to_child_path(path);
            path =  list_filter.convert_path_to_child_path(path);
            list_store.get_iter (out iter, path);
            list_store.get (iter, 2, out app);
            selected_app = app;
        }

        if (command.text != "" && command.text != null)
            create_command ();
    }

    // The filter, searches the application name and description
    private bool filter_func (Gtk.TreeModel model, Gtk.TreeIter iter) {
        AppInfo app_info;
        string name;

        list_store.get (iter, 2, out app_info);
        name =  app_info.get_name () + app_info.get_description ();

        if (search_text == "")
            return true;
        else if (name != null)
            return name.up ().contains (search_text.up ());
        else
            return false;
    }

    // Create custom AppInfo
    private void create_command () {
        if (command.text != "" || command.text != null) {
            var keyfile = new KeyFile ();
            keyfile.set_string ("Desktop Entry", "Name", _("Custom Command"));
            keyfile.set_string ("Desktop Entry", "GenericName", command.text);
            keyfile.set_string ("Desktop Entry", "Comment", command.text);
            keyfile.set_string ("Desktop Entry", "Exec", command.text);
            keyfile.set_string ("Desktop Entry", "Icon", FALLBACK_ICON);
            keyfile.set_string ("Desktop Entry", "Type", "Application");

            selected_app = new DesktopAppInfo.from_keyfile (keyfile) as AppInfo;
        }
    }
}
