//  
//  Copyright (C) 2013 Michael Langfermann
// 
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
// 

public class StartupPlug : Pantheon.Switchboard.Plug {
    private Gtk.ListStore list_store;
    private Gtk.TreeModelSort list_sort;
    private Gtk.TreeView list;
    private Gtk.CellRendererText cell;
    private Gtk.ToolButton add_button;
    private Gtk.ToolButton remove_button;
    private Gtk.ToolButton config_button;

    private const int ICON_SIZE = 48;
    private const string FALLBACK_ICON = "application-default-icon";

    private string list_markup = "<span font_weight=\"bold\" size=\"large\">{0}</span>\n{1}";

    public StartupPlug () {
        create_widgets ();
        connect_signals ();
        list_update();
    }

    private void create_widgets () {
        // Create ListStore
        list_store = new Gtk.ListStore (4,
                                        typeof (string),         // Application name and description
                                        typeof (Gdk.Pixbuf),     // Application icon
                                        typeof (string),         // File path
                                        typeof (DesktopAppInfo));// Hidden returned application data

         // Create a sort model wrapper, sort in ascending order
        list_sort = new Gtk.TreeModelSort.with_model (list_store);
        list_sort.set_sort_column_id (0, Gtk.SortType.ASCENDING);

        // Create a TreeView with the sorted model
        list = new Gtk.TreeView.with_model (list_sort);
        list.expand = true;
        list.enable_search = false;
        list.headers_visible = false;

        // Create and append columns
        Gtk.CellRendererPixbuf pixbuf = new Gtk.CellRendererPixbuf ();
        list.insert_column_with_attributes (-1, null, pixbuf, "pixbuf", 1);

        cell = new Gtk.CellRendererText ();
        list.insert_column_with_attributes (-1, null, cell, "markup", 0);

        // Layout widgets
        var scroll = new Gtk.ScrolledWindow (null, null);
        scroll.hscrollbar_policy = Gtk.PolicyType.NEVER;
        scroll.vscrollbar_policy = Gtk.PolicyType.AUTOMATIC;
        scroll.shadow_type = Gtk.ShadowType.IN;
        scroll.expand = true;

        scroll.add (list);

        // Toolbar
        var tbar = new Gtk.Toolbar ();
        tbar.set_style(Gtk.ToolbarStyle.ICONS);
        tbar.set_icon_size(Gtk.IconSize.SMALL_TOOLBAR);
        tbar.set_show_arrow (false);
        tbar.hexpand = true;

        tbar.get_style_context ().add_class (Gtk.STYLE_CLASS_INLINE_TOOLBAR);
        tbar.get_style_context ().set_junction_sides (Gtk.JunctionSides.TOP);

        add_button = new Gtk.ToolButton (null, _("Add..."));
        remove_button = new Gtk.ToolButton (null, _("Remove"));
        config_button = new Gtk.ToolButton (null, _("Configure Startup Application"));
        var spacer = new Gtk.ToolItem ();

        spacer.set_expand(true);

        add_button.set_icon_name ("list-add-symbolic");
        remove_button.set_icon_name ("list-remove-symbolic");
        config_button.set_icon_name ("document-properties-symbolic");

        add_button.set_tooltip_text (_("Add..."));
        remove_button.set_tooltip_text (_("Remove"));
        config_button.set_tooltip_text (_("Configure Startup Application"));

        tbar.insert (add_button, -1);
        tbar.insert (remove_button, -1);
        tbar.insert (spacer, -1);
        tbar.insert (config_button, -1);;

        // The Box
        var box = new Gtk.Box (Gtk.Orientation.VERTICAL, 0);

        box.pack_start (scroll, true);
        box.pack_start (tbar, false);

        add (box);
    }

    // Load applications into ListStore
    private void list_update () {
        Gtk.TreeIter iter;
        FileInfo file_info;
        var icon_theme = Gtk.IconTheme.get_default ();
        var dir = Environment.get_user_config_dir () + "/autostart/";
        list_store.clear ();

        try {
            if (!FileUtils.test (dir, FileTest.EXISTS))
                DirUtils.create (dir, 0x1ED);
            var enumerator = File.new_for_path (dir)
                                 .enumerate_children (FileAttribute.STANDARD_NAME, 0);

            while ((file_info = enumerator.next_file ()) != null) {
                var name = file_info.get_name ();
                if (!name.contains ("~") && name.contains (".desktop")) {
                    var app = new DesktopAppInfo.from_filename (dir + name);

                    var app_name = app.get_name ();
                    var app_desc = app.get_description ();

                    (app_desc == null)? app_desc = "":null;

                    list_store.append (out iter);
                    list_store.set (iter, 0, list_markup.replace("{0}", app_name)
                                                        .replace("{1}", app_desc));

                    if (Gtk.IconTheme.get_default ().has_icon (app.get_icon ().to_string ()))
                        list_store.set (iter, 1,
                                        icon_theme.load_icon (app.get_icon ().to_string (),
                                        ICON_SIZE, Gtk.IconLookupFlags.GENERIC_FALLBACK));
                    else
                        list_store.set (iter, 1,
                                        icon_theme.load_icon (FALLBACK_ICON,
                                        ICON_SIZE, Gtk.IconLookupFlags.GENERIC_FALLBACK));

                    list_store.set (iter, 2, dir + name);
                    list_store.set (iter, 3, app);
                }
            }
        } catch (Error e) { warning (e.message); }
    }

    // Connect signals
    private void connect_signals () {

        add_button.clicked.connect (() => {
            var dialog = new AppDialog();
            dialog.custom_command = true;
            dialog.modal = true;

            void* pointer;
            this.socket_window.get_user_data(out pointer);

            dialog.transient_for = ((Gtk.Window)pointer);

            dialog.app_chosen.connect (() => {
                add_app (dialog.selected_app);
                list_update ();
            });

            dialog.show_all ();
            dialog.run ();
            dialog.destroy ();
        });

        remove_button.clicked.connect (() => app_remove ());

        config_button.clicked.connect (() => {
            var app_info = get_app();
            if (app_info != null) {
                var popover = new Granite.Widgets.PopOver();
                var entry_name = new Gtk.Entry ();
                var entry_desc = new Gtk.Entry ();
                var entry_command = new Gtk.Entry ();
                var apply = new Gtk.ToolButton.from_stock (Gtk.Stock.EDIT);
                var grid = new Gtk.Grid ();

                entry_name.text = app_info.get_name ();
                entry_desc.text = app_info.get_description ();
                entry_command.text = app_info.get_commandline ();
                entry_name.width_request = 160;
                entry_desc.width_request = 160;
                entry_command.width_request = 160;
                apply.halign = Gtk.Align.START;

                grid.row_spacing = 6;
                grid.column_spacing = 12;

                apply.clicked.connect (() => {
                    app_edit(entry_name.text, entry_desc.text, entry_command.text);
                    popover.destroy ();
                });

                entry_name.activate.connect (() => apply.clicked());
                entry_desc.activate.connect (() => apply.clicked());
                entry_command.activate.connect (() => apply.clicked());

                grid.attach (new LLabel.right (_("Name:")), 0, 0, 1, 1);
                grid.attach (entry_name, 1, 0, 1, 1);

                grid.attach (new LLabel.right (_("Description:")), 0, 1, 1, 1);
                grid.attach (entry_desc, 1, 1, 1, 1);

                grid.attach (new LLabel.right (_("Command:")), 0, 2, 1, 1);
                grid.attach (entry_command, 1, 2, 1, 1);

                grid.attach (apply, 2, 2, 1, 1);

                ((Gtk.Box)popover.get_content_area ()).add (grid);
                popover.move_to_widget(config_button);
                popover.show_all ();
                popover.present ();
                popover.run ();
                popover.destroy ();
            }
        });
    }

    // Add Startup Item
    private void add_app (AppInfo selected_app) {
        var keyfile = new KeyFile ();
        var file_path = Environment.get_user_config_dir () + "/autostart/";
        var file_name = selected_app.get_name ();
        var file_strip = file_name.strip ().down ();
        int i = 1;

        try {
            if (FileUtils.test (file_path + file_strip + ".desktop", FileTest.EXISTS)) {
                i++;
                while (FileUtils.test (file_path + file_strip + i.to_string () + ".desktop",
                                       FileTest.EXISTS))
                       i++;
                file_name += " " + i.to_string ();
                file_strip += i.to_string ();
            }

            keyfile.set_string ("Desktop Entry", "Name", file_name);
            keyfile.set_string ("Desktop Entry", "GenericName", selected_app.get_display_name ());
            keyfile.set_string ("Desktop Entry", "Comment", selected_app.get_description ());
            keyfile.set_string ("Desktop Entry", "Exec", selected_app.get_commandline ());
            keyfile.set_string ("Desktop Entry", "Icon", selected_app.get_icon ().to_string ());
            keyfile.set_string ("Desktop Entry", "Type", "Application");

            FileUtils.set_contents (file_path + 
                                    file_strip + ".desktop",
                                    keyfile.to_data(null));
        } catch (Error e) { warning (e.message); }
    }

    // Removes the .desktop file
    private void app_remove() {
        GLib.Value name;
        Gtk.TreeIter iter;
        Gtk.TreePath path;

        list.get_cursor (out path, null);
        if (path != null) {
            path =  list_sort.convert_path_to_child_path(path);
            list_store.get_iter (out iter, path);
            list_store.get_value (iter, 2, out name);
            FileUtils.remove (name.get_string ());
        }

        list_update ();
    }

    // Edit an Item
    private void app_edit(string name, string desc, string command) {
        AppInfo app;
        Gtk.TreeIter iter;
        Gtk.TreePath path;
        string file_path;
        var keyfile = new KeyFile ();

        list.get_cursor (out path, null);
        if (path != null) {
            path =  list_sort.convert_path_to_child_path(path);
            list_store.get_iter (out iter, path);
            list_store.get (iter, 3, out app);
            list_store.get (iter, 2, out file_path);

            try {
                keyfile.load_from_file(file_path, KeyFileFlags.KEEP_COMMENTS | KeyFileFlags.KEEP_TRANSLATIONS);

                keyfile.set_string ("Desktop Entry", "Name", name);
                keyfile.set_string ("Desktop Entry", "Comment", desc);
                keyfile.set_string ("Desktop Entry", "Exec", command);

                FileUtils.set_contents (file_path, keyfile.to_data(null));
            } catch (Error e) { warning (e.message); }

        }
        list_update ();
    }

    // Gets the selected app and returns it as a DesktopAppInfo object
    private DesktopAppInfo? get_app() {
        DesktopAppInfo app;
        Gtk.TreeIter iter;
        Gtk.TreePath path;

        list.get_cursor (out path, null);

        if (path != null) {
            path =  list_sort.convert_path_to_child_path(path);
            list_store.get_iter (out iter, path);
            list_store.get (iter, 3, out app);
            return app;
        }

        return null;
    }
}

public static int main (string[] args) {
    Gtk.init (ref args);

    var plug = new StartupPlug ();
    plug.register ("Startup");
    plug.show_all ();

    Gtk.main ();
    return 0;
}